# -*- coding: utf-8 -*-
"""
Created on Mon Oct 27 17:48:33 2014

@author: akusok
"""

import numpy as np


def train_v(self, X, T, Xv, Tv):
    HH, HT, Beta = self._project(X, T, solve=True)
    Hv = self.project(Xv)
    nn = Hv.shape[1]
    e = np.ones((nn+1,)) * -1  # errors for all numbers of neurons
    rank, nn = self._ranking(nn, Xv, Tv)  # create ranking of neurons

    Yv = np.dot(Hv, Beta)
    err = self._error(Yv, Tv)
    penalty = err * 0.01 / nn  # penalty is 1% of error at max(nn)
    e[nn] = err + nn * penalty

    # MYOPT function
    # [A  B  C  D  E] interval points,
    # halve the interval each time

    # initialize intervals
    A = 1
    E = nn
    l = E - A
    B = A + l/4
    C = A + l/2
    D = A + 3*l/4

    l = 3  # run the while loop at least once
    while l > 2:
        # calculate errors at points
        for idx in [A, B, C, D, E]:
            if e[idx] == -1:  # skip already calculated errors
                rank1 = rank[:idx]
                HH1 = HH[rank1, :][:, rank1]
                HT1 = HT[rank1, :]
                Beta = self._solve_corr(HH1, HT1)
                Yv = np.dot(Hv[:, rank1], Beta)
                e[idx] = self._error(Yv, Tv) + idx * penalty

        m = min(e[A], e[B], e[C], e[D], e[E])  # find minimum element

        # halve the search interval
        if m in (e[A], e[B]):
            E = C
            C = B
        elif m in (e[D], e[E]):
            A = C
            C = D
        else:
            A = B
            E = D
        l = E - A
        B = A + l/4
        D = A + (3*l)/4

    k_opt = [n1 for n1 in [A, B, C, D, E] if e[n1] == m][0]  # find minimum index
    best_nn = rank[:k_opt]

    self._prune(best_nn)
    self.Beta = self._project(X, T, solve=True)[2]
#    print "%d of %d neurons selected with a validation set" % (len(best_nn), nn)
#    if len(best_nn) > nn*0.9:
#        print "Hint: try re-training with more hidden neurons"





















# -*- coding: utf-8 -*-
"""Multiresponse Sparse Regression algorithm in Python

Uses np library from Python

| Input:
| **T**    is an (n x p) matrix of targets. The columns of T should
      have zero mean and same scale (e.g. equal variance).
| **X**    is an (n x m) matrix of regressors. The columns of X should
      have zero mean and same scale (e.g. equal variance).
| **kmax** is an integer fixing the number of steps to be run, which
      equals to the maximum number of regressors in the model.
  
| Output:
| **W**    is an (m x p*kmax) sparse matrix of regression
      coefficients. It can be converted to full matrix by command   
      full(W). Regression coefficients of the k:th step are given
      by W(:,(k-1)*p+1:k*p).
| **i1**   is a (1 x kmax) vector of indices revealing the order in
      which the regressors enter model. 
 
The estimates for T may be obtained by Y = X*W, where the k:th
estimate Y(:,(k-1)*p+1:k*p) uses k regressors.
  
| Reference: 
| Timo Simila, Jarkko Tikka. Multiresponse sparse regression with
 application to multidimensional scaling. International Conference
 on Artificial Neural Networks (ICANN). Warsaw, Poland. September
 11-15, 2005. LNCS 3697, pp. 97-102.

Copyright (C) 2005 by Timo Simila and Jarkko Tikka.

This function is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or any later version.   

The function is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the `GNU
General Public License  <http://www.gnu.org/copyleft/gpl.html>`_  
for more details.
"""

import numpy as np


def mrsr(X, T, kmax):

    n,m = X.shape
    n,p = T.shape
    kmax = min(kmax, m)
    if p > 15:
        print "Too many targets (%d) - MRSR has O(2^targets) complexity"
    """    print "Reducing to 15 randomly selected targets (6x slowdown)"
        print "Using max 10 targets (1.08x slowdown) recommended"
        ti = np.arange(p)
        np.random.shuffle(ti)
        ti = ti[:15]
        T = T[:,ti]
        p = 15
    """
    i1 = np.array([], dtype = np.int32)
    i2 = np.arange(m).astype(np.int32)
    XT = np.dot(X.T,T)
    XX = np.zeros([m, m])
	
    S = np.ones([2**p, p])
    S[0:2**(p-1), 0] = -1
    for j in np.arange(1, p):
        S[:, j] = np.concatenate((S[np.arange(1, 2**p, 2), j-1], S[np.arange(1, 2**p, 2), j-1]))
	
	
	
    # Make the first step
	
    A     = np.transpose(XT)
    cmax  = np.amax(abs(A).sum(0), 0)
    cind  = np.argmax(abs(A).sum(0), 0)
    A     = np.delete(A, cind, 1)
    ind   = int(i2[cind])
    i2    = np.delete(i2, cind)
    i1    = np.append(i1, ind)
    # here Xi1 and Xi2 are just faster alternatives to X[:,i1] and X[:,i2]
    Xi2   = X.copy(order='F')  # column-contiguous copy of X
    Xi2[:, cind:-1] = Xi2[:, cind+1:];  Xi2   = Xi2[:,:-1]  # delete <cind> col
    Xi1   = X[:,ind].reshape((n,1))  # add 1 column at a time

    XX[np.ix_([ind], [ind])] = np.dot(X[:,ind], X[:,ind])
    
    invXX = 1 / XX[ind, :][ind]
    Wols  = invXX * XT[ind, :]
    Yols  = np.dot(Xi1, np.reshape(Wols, (1,-1)))
    B     = np.dot(Yols.T, Xi2)
    G     = (cmax+np.dot(S,A))/(cmax+np.dot(S,B))
    g     = G[G>=0].min()
    Y = g*Yols
	
    # Rest of the steps
    for k in np.arange(2,kmax+1):
        #print "calculating rank %d/%d" % (k-1, kmax)		
        #print "mrsr %d/%d" % (k+1, kmax)
  
        A    = np.dot((T-Y).T, Xi2)  # true slow
        cmax = np.amax(abs(A).sum(0), 0)
        cind = np.argmax(abs(A).sum(0), 0)
        A    = np.delete(A, cind, 1)
        ind  = int(i2[cind])
        i2   = np.delete(i2, cind)
        i1   = np.append(i1, ind)
        #Xi1  = np.hstack((Xi1, X[:,ind].reshape((n,1), order='C')))  # slow for large k
        Xi1  = np.hstack((Xi1, X[:,ind].reshape((-1,1))))  # slow for large k
        xX   = np.dot(X[:, ind].T, Xi1)
		
        XX[np.ix_([ind], i1)] = xX
        XX[np.ix_(i1, [ind])] = np.reshape(xX, (i1.size, -1))
	
        v3 = XX.take(i1,axis=0).take(i1,axis=1)  # XX[i1, :][:, i1]
        #v3 = XX[i1, :][:, i1]
        try:
            invXX = np.linalg.inv(v3)	   
        except np.linalg.linalg.LinAlgError:
            print 'got singular matrix, using pinv()', i1
            invXX = np.linalg.pinv(v3)
    
        Wols = np.dot(invXX, XT.take(i1,axis=0))
        Yols = np.dot(Xi1, Wols)  # true slow
        # deletes [cind] row, slow for large k
        Xi2[:, cind:-1] = Xi2[:, cind+1:];  Xi2 = Xi2[:,:-1]
        
        B    = np.dot((Yols-Y).T, Xi2)  # true slow
        
        G    = (cmax + S.dot(A)) / (cmax + S.dot(B))  # true slow for many outputs
        # now we remove that line using a condition:
        # G = numpy.concatenate(([2*(k==m)-1], G.flatten()), 1)
        if k == kmax:  # G[G>=0] is empty if k==kmax; empty.min() will give error
            Y = Yols
        else:
            g = G[G>=0].min()
            Y = (1-g)*Y+g*Yols
		
    return i1
	



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 27 17:48:33 2014

@author: akusok
"""

import numpy as np
import numexpr as ne
from tables import open_file
from scipy.spatial.distance import cdist
from scipy.linalg import solve as cpu_solve
from multiprocessing import Pool, cpu_count
import cPickle
import os, platform


def cd(a):
    x, w, kind, idx = a
    return cdist(x, w, kind)**2, idx


class SLFN(object):
    """Single-hidden Layer Feed-forward Network.
    """

    def __init__(self, inputs, targets, batch=100, accelerator=""):
        """Initializes a SLFN with an empty hidden layer.

        :param inputs: number of features in input samples (input dimensionality)
        :param outputs: number of simultaneous predicted outputs
        """
        assert isinstance(inputs, (int, long)), "Number of inputs must be integer"
        assert isinstance(targets, (int, long)), "Number of outputs must be integer"
        assert batch > 0, "Batch should be positive"

        self.inputs = inputs
        self.targets = targets
        # cannot use a dictionary for neurons, because its iteration order is not defined
        self.neurons = []  # list of all neurons with their types (= transformantion functions)
        self.Beta = None
        self.flist = ("lin", "sigm", "tanh", "rbf_l1", "rbf_l2", "rbf_linf")
        self.alpha = 1E-9  # normalization for H'H solution
        self.batch = int(batch)  # smallest batch for batch processing
        self.accelerator = None  # None, "GPU", "PHI"
        if accelerator == "GPU":
            self.accelerator = "GPU"
            self.magma_solver = __import__('acc.magma_solver', globals(), locals(), ['gpu_solve'], -1)
            print "using GPU"
        # init other stuff
        self.opened_hdf5 = []
        self.classification = None  # c / wc / mc
        self.weights_wc = None
        self.tprint = 5

    def __del__(self):
        """Close HDF5 files opened during HPELM usage.
        """
        if len(self.opened_hdf5) > 0:
            for h5 in self.opened_hdf5:
                h5.close()

    def _checkdata(self, X, T):
        """Checks data variables and fixes matrix dimensionality issues.
        """
        if X is not None:
            if isinstance(X, basestring):  # open HDF5 file
                try:
                    h5 = open_file(X, "r")
                    self.opened_hdf5.append(h5)
                    for node in h5.walk_nodes():
                        pass  # find a node with whatever name
                    X = node
                except:
                    raise IOError("Cannot read HDF5 file at %s" % X)
            else:
                # assert isinstance(X, np.ndarray) and X.dtype.kind not in "OSU", "X must be a numerical numpy array"
                if len(X.shape) == 1:
                    X = X.reshape(-1, 1)
            assert len(X.shape) == 2, "X must have 2 dimensions"
            assert X.shape[1] == self.inputs, "X has wrong dimensionality: expected %d, found %d" % (self.inputs, X.shape[1])

        if T is not None:
            if isinstance(T, basestring):  # open HDF5 file
                try:
                    h5 = open_file(T, "r")
                    self.opened_hdf5.append(h5)
                    for node in h5.walk_nodes():
                        pass  # find a node with whatever name
                    T = node
                except:
                    raise IOError("Cannot read HDF5 file at %s" % T)
            else:
                # assert isinstance(T, np.ndarray) and T.dtype.kind not in "OSU", "T must be a numerical numpy array"
                if len(T.shape) == 1:
                    T = T.reshape(-1, 1)
            assert len(T.shape) == 2, "T must have 2 dimensions"
            assert T.shape[1] == self.targets, "T has wrong dimensionality: expected %d, found %d" % (self.targets, T.shape[1])

        if (X is not None) and (T is not None):
            assert X.shape[0] == T.shape[0], "X and T cannot have different number of samples"

        return X, T

    def add_neurons(self, number, func, W=None, B=None):
        """Add neurons of a specific type to the SLFN.

        If neurons of such type exist, merges them together.
        :param number: - number of neurons to add
        :param func: - transformation function of those neurons,
                       "lin", "sigm", "tanh", "rbf_l1", "rbf_l2", "rbf_linf"
                       or a custom function of type <numpy.ufunc>
        :param W: - projection matrix or ("rbf_xx") a list of centroids
        :param B: - bias vector or ("rbf_xx") a vector of sigmas
        """
        assert isinstance(number, int), "Number of neurons must be integer"
        assert func in self.flist or isinstance(func, np.ufunc), "Use standard neuron function or a custom <numpy.ufunc>"
        assert isinstance(W, (np.ndarray, type(None))), "Projection matrix (W) must be a Numpy ndarray"
        assert isinstance(B, (np.ndarray, type(None))), "Bias vector (B) must be a Numpy ndarray"

        # initialize skipped arguments
        if W is None:
            if func == "lin":  # copying input features for linear neurons
                number = min(number, self.inputs)  # cannot have more linear neurons than features
                W = np.eye(self.inputs, number)
            else:
                W = np.random.randn(self.inputs, number)
                if func not in ("rbf_l1", "rbf_l2", "rbf_linf"):
                    W = W * (3 / self.inputs ** 0.5)  # high dimensionality fix
        if B is None:
            B = np.random.randn(number)
            if func in ("rbf_l2", "rbf_l1", "rbf_linf"):
                B = np.abs(B)
                B = B * self.inputs
            if func == "lin":
                B = np.zeros((number,))
        assert W.shape == (self.inputs, number), "W must be size [inputs, neurons] (expected [%d,%d])" % (self.inputs, number)
        assert B.shape == (number,), "B must be size [neurons] (expected [%d])" % number

        ntypes = [nr[0] for nr in self.neurons]  # existing types of neurons
        if func in ntypes:
            # add to an existing neuron type
            i = ntypes.index(func)
            _, nn0, W0, B0 = self.neurons[i]
            number = nn0 + number
            W = np.hstack((W0, W))
            B = np.hstack((B0, B))
            self.neurons[i] = (func, number, W, B)
        else:
            # create a new neuron type
            self.neurons.append((func, number, W, B))
        self.Beta = None  # need to re-train network after adding neurons


    def project(self, X):
        # assemble hidden layer output with all kinds of neurons
        assert len(self.neurons) > 0, "Model must have hidden neurons"

        X, _ = self._checkdata(X, None)
        H = []
        cdkinds = {"rbf_l2": "euclidean", "rbf_l1": "cityblock", "rbf_linf": "chebyshev"}
        for func, _, W, B in self.neurons:
            # projection
            if "rbf" in func:
                self._affinityfix()
                N = X.shape[0]
                k = cpu_count()
                jobs = [(X[idx], W.T, cdkinds[func], idx) for idx in np.array_split(np.arange(N), k*10)]  #### ERROR HERE!!!
                p = Pool(k)
                H0 = np.empty((N, W.shape[1]))
                for h0, idx in p.imap(cd, jobs):
                    H0[idx] = h0
                p.close()
                H0 = - H0 / B
#            if func == "rbf_l2":
#                H0 = - cdist(X, W.T, "euclidean")**2 / B
#            elif func == "rbf_l1":
#                H0 = - cdist(X, W.T, "cityblock")**2 / B
#            elif func == "rbf_linf":
#                H0 = - cdist(X, W.T, "chebyshev")**2 / B
            else:
                H0 = X.dot(W) + B

            # transformation
            if func == "lin":
                pass
            elif "rbf" in func:
                ne.evaluate('exp(H0)', out=H0)
            elif func == "sigm":
                ne.evaluate("1/(1+exp(-H0))", out=H0)
            elif func == "tanh":
                ne.evaluate('tanh(H0)', out=H0)
            else:
                H0 = func(H0)  # custom <numpy.ufunc>
            H.append(H0)

        if len(H) == 1:
            H = H[0]
        else:
            H = np.hstack(H)
#        print (H > 0.01).sum(0)
        return H

    def predict(self, X):
        """Predict targets for the given inputs X.

        :param X: - model inputs
        """
        assert self.Beta is not None, "Train ELM before predicting"
        H = self.project(X)
        Y = H.dot(self.Beta)
        return Y

    def error(self, Y, T):
        """Calculate error of model predictions.
        """
        _, Y = self._checkdata(None, Y)
        _, T = self._checkdata(None, T)
        return self._error(Y, T)

    def confusion(self, Y1, T1):
        """Compute confusion matrix for the given classification, iteratively.
        """
        _, Y = self._checkdata(None, Y1)
        _, T = self._checkdata(None, T1)
        nn = np.sum([n1[1] for n1 in self.neurons])
        N = T.shape[0]
        batch = max(self.batch, nn)
        nb = N / batch  # number of batches
        if batch > N * nb:
            nb += 1

        C = self.targets
        conf = np.zeros((C, C))

        if self.classification in ("c", "wc"):
            for b in xrange(nb):
                start = b*batch
                stop = min((b+1)*batch, N)
                Tb = np.array(T[start:stop]).argmax(1)
                Yb = np.array(Y[start:stop]).argmax(1)
                for c1 in xrange(C):
                    for c1h in xrange(C):
                        conf[c1, c1h] += np.sum((Tb == c1) * (Yb == c1h))
        elif self.classification == "mc":
            for b in xrange(nb):
                start = b*batch
                stop = min((b+1)*batch, N)
                Tb = np.array(T[start:stop]) > 0.5
                Yb = np.array(Y[start:stop]) > 0.5
                for c1 in xrange(C):
                    for c1h in xrange(C):
                        conf[c1, c1h] += np.sum(Tb[:, c1] * Yb[:, c1h])
        else:  # No confusion matrix
            conf = None
        return conf

    ######################
    ### helper methods ###

    def _prune(self, idx):
        """Leave only neurons with the given indexes.
        """
        idx = list(idx)
        neurons = []
        for nold in self.neurons:
            k = nold[1]  # number of neurons
            ix1 = [i for i in idx if i < k]  # index for current neuron type
            idx = [i-k for i in idx if i >= k]
            func = nold[0]
            number = len(ix1)
            W = nold[2][:, ix1]
            bias = nold[3][ix1]
            neurons.append((func, number, W, bias))
        self.neurons = neurons

    def _ranking(self, nn):
        """Returns a random ranking of hidden neurons.
        """
        rank = np.arange(nn)
        np.random.shuffle(rank)
        return rank, nn

    def _solve_corr(self, HH, HT):
        """Solve a linear system from correlation matrices.
        """
        if self.accelerator == "GPU":
            Beta = self.magma_solver.gpu_solve(HH, HT, self.alpha)
        else:
            Beta = cpu_solve(HH, HT, sym_pos=True)
        return Beta

    def _error(self, Y, T, R=None):
        """Returns regression/classification/multiclass error, also for PRESS.
        """
        raise NotImplementedError("SLFN does not know the use case to compute an error")

    def _train(self, X, T):
        raise NotImplementedError("SLFN does not know the use case to train a network")

    def __str__(self):
        s = "SLFN with %d inputs and %d outputs\n" % (self.inputs, self.targets)
        s += "Hidden layer neurons: "
        for func, n, _, _ in self.neurons:
            s += "%d %s, " % (n, func)
        s = s[:-2]
        return s

    def _affinityfix(self):
        # Numpy processor affinity fix
        if "Linux" in platform.system():
            a = np.random.rand(3, 1)
            np.dot(a.T, a)
            pid = os.getpid()
            os.system("taskset -p 0xffffffff %d >/dev/null" % pid)

    def save(self, fname):
        assert isinstance(fname, basestring), "Model file name must be a string"
        m = {"inputs": self.inputs,
             "outputs": self.targets,
             "neurons": self.neurons,
             "Beta": self.Beta,
             "alpha": self.alpha,
             "Classification": self.classification,
             "Weights_WC": self.weights_wc}
        try:
            cPickle.dump(m, open(fname, "wb"), -1)
        except IOError:
            raise IOError("Cannot create a model file at: %s" % fname)

    def load(self, fname):
        assert isinstance(fname, basestring), "Model file name must be a string"
        try:
            m = cPickle.load(open(fname, "rb"))
        except IOError:
            raise IOError("Model file not found: %s" % fname)
        self.inputs = m["inputs"]
        self.targets = m["outputs"]
        self.neurons = m["neurons"]
        self.Beta = m["Beta"]
        self.alpha = m["alpha"]
        self.classification = m["Classification"]
        self.weights_wc = m["Weights_WC"]





























# -*- coding: utf-8 -*-
"""
Created on Fri Mar 20 14:27:13 2015

@author: akusok
"""

from gpu_solver import GPUSolver, gpu_solve
import numpy as np
from scipy.linalg import solve
import sys
from time import time


def s_cpu(data, n, nrhs):
    XX = np.zeros((n, n))
    XT = np.zeros((n, nrhs))
    for X, T in data:
        t = time()
        XX += X.T.dot(X)
        XT += X.T.dot(T)
        print "CPU: added batch in %.2fs" % (time()-t)
    t = time()
    Beta = solve(XX, XT, sym_pos=True)
    print "CPU: solution in %.2fs" % (time()-t)
    return Beta


def s_gpu(data, n, nrhs):
    s = GPUSolver(n, nrhs)
    for X, T in data:
        t = time()
        s.add_data(X, T)
        print "GPU: added batch in %.2fs" % (time()-t)
    t = time()
    B = s.solve()
    print "GPU: solution in %.2fs" % (time()-t)
    return B


def test():
    n = int(sys.argv[1])
    nrhs = int(sys.argv[2])
    k = int(sys.argv[3])

    data = []
    for _ in range(k):
        n1 = int(n*0.7)
        X = np.random.rand(n1, n).astype(np.float64)
        T = np.random.rand(n1, nrhs).astype(np.float64)
        data.append((X, T))

    B = s_gpu(data, n, nrhs)
    Beta = s_cpu(data, n, nrhs)

    print "Norm of CPU-GPU difference:", np.linalg.norm(B - Beta)


if __name__ == "__main__":
    
    if len(sys.argv) != 4:
        print "add arguments: number of neurons, number of targets, number of batches of data"
    else:
        test()
    
    print "Works!"
























































# -*- coding: utf-8 -*-
"""
Created on Mon Oct 27 17:48:33 2014

@author: akusok
"""

import numpy as np
from numpy.linalg import inv


def train_loo(self, X, T):
    H = self.project(X)
    HH, HT = self._project(X, T)

    N, nn = H.shape
    e = np.ones((nn+1,)) * -1  # errors for all numbers of neurons
    rank, nn = self._ranking(nn, H, T)  # create ranking of neurons

    # PRESS_LOO
    P = inv(HH)
    Bt = np.dot(P, HT)
    R = np.ones((N,)) - np.einsum('ij,ji->i', np.dot(H, P), H.T)
    Y = np.dot(H, Bt)
    err = self._error(Y, T, R)

    penalty = err * 0.01 / nn  # penalty is 1% of error at max(nn)
    e[nn] = err + nn * penalty

    # MYOPT function
    # [A  B  C  D  E] interval points,
    # halve the interval each time

    # initialize intervals
    A = 1
    E = nn
    l = E - A
    B = A + l/4
    C = A + l/2
    D = A + 3*l/4

    l = 1000  # run the while loop at least once
    while l > 2:
        # calculate errors at points
        for idx in [A, B, C, D, E]:
            if e[idx] == -1:  # skip already calculated errors
                rank1 = rank[:idx]
                # H1 = H[:, rank1]
                H1 = np.take(H, rank1, 1)
                HH1 = HH[rank1, :][:, rank1]
                HT1 = HT[rank1, :]

                P = inv(HH1)
                Bt1 = np.dot(P, HT1)
                R = np.ones((N,)) - np.einsum('ij,ji->i', np.dot(H1, P), H1.T)
                Y1 = np.dot(H1, Bt1)
                err = self._error(Y1, T, R)
                e[idx] = err + idx * penalty

        m = min(e[A], e[B], e[C], e[D], e[E])  # find minimum element

        # halve the search interval
        if m in (e[A], e[B]):
            E = C
            C = B
        elif m in (e[D], e[E]):
            A = C
            C = D
        else:
            A = B
            E = D
        l = E - A
        B = A + l/4
        D = A + (3*l)/4

    k_opt = [n1 for n1 in [A, B, C, D, E] if e[n1] == m][0]  # find minimum index
    best_nn = rank[:k_opt]

    self._prune(best_nn)
    self.Beta = self._project(X, T, solve=True)[2]
#    print "%d of %d neurons selected with a LOO validation" % (len(best_nn), nn)
#    if len(best_nn) > nn*0.9:
#        print "Hint: try re-training with more hidden neurons"


































